@@ -6,9 +6,9 @@
 /// by the user, it is determined by the classification of the syntax theme into light-background
 /// vs dark-background syntax themes. If the user didn't choose a syntax theme, a dark-background
 /// default is selected.
-use syntect::highlighting::ThemeSet;
+use bat;
+use bat::assets::HighlightingAssets;
 
-use crate::bat_utils::assets::HighlightingAssets;
 use crate::cli;
 use crate::env;
 
@@ -22,16 +22,15 @@ pub fn set__is_light_mode__syntax_theme__syntax_set(
         opt.syntax_theme.as_ref(),
         syntax_theme_name_from_bat_theme.as_ref(),
         opt.light,
-        &assets.theme_set,
     );
     opt.computed.is_light_mode = is_light_mode;
 
     opt.computed.syntax_theme = if is_no_syntax_highlighting_syntax_theme_name(&syntax_theme_name) {
         None
     } else {
-        Some(assets.theme_set.themes[&syntax_theme_name].clone())
+        Some(assets.get_theme(&syntax_theme_name).clone())
     };
-    opt.computed.syntax_set = assets.syntax_set;
+    opt.computed.syntax_set = assets.get_syntax_set().unwrap().clone();
 }
 
 pub fn is_light_syntax_theme(theme: &str) -> bool {
@@ -86,45 +85,26 @@ fn get_is_light_mode_and_syntax_theme_name(
     theme_arg: Option<&String>,
     bat_theme_env_var: Option<&String>,
     light_mode_arg: bool,
-    theme_set: &ThemeSet,
 ) -> (bool, String) {
-    let theme_arg = valid_syntax_theme_name_or_none(theme_arg, theme_set);
-    let bat_theme_env_var = valid_syntax_theme_name_or_none(bat_theme_env_var, theme_set);
     match (theme_arg, bat_theme_env_var, light_mode_arg) {
         (None, None, false) => (false, DEFAULT_DARK_SYNTAX_THEME.to_string()),
-        (Some(theme_name), _, false) => (is_light_syntax_theme(&theme_name), theme_name),
-        (None, Some(theme_name), false) => (is_light_syntax_theme(&theme_name), theme_name),
-        (None, None, true) => (true, DEFAULT_LIGHT_SYNTAX_THEME.to_string()),
-        (Some(theme_name), _, is_light_mode) => (is_light_mode, theme_name),
-        (None, Some(theme_name), is_light_mode) => (is_light_mode, theme_name),
-    }
-}
-
-// At this stage the theme name is considered valid if it is either a real theme name or the special
-// no-syntax-highlighting name.
-fn valid_syntax_theme_name_or_none(
-    theme_name: Option<&String>,
-    theme_set: &ThemeSet,
-) -> Option<String> {
-    match theme_name {
-        Some(name)
-            if is_no_syntax_highlighting_syntax_theme_name(name)
-                || theme_set.themes.contains_key(name) =>
-        {
-            Some(name.to_string())
+        (Some(theme_name), _, false) => (is_light_syntax_theme(theme_name), theme_name.to_string()),
+        (None, Some(theme_name), false) => {
+            (is_light_syntax_theme(theme_name), theme_name.to_string())
         }
-        _ => None,
+        (None, None, true) => (true, DEFAULT_LIGHT_SYNTAX_THEME.to_string()),
+        (Some(theme_name), _, is_light_mode) => (is_light_mode, theme_name.to_string()),
+        (None, Some(theme_name), is_light_mode) => (is_light_mode, theme_name.to_string()),
     }
 }
 
 #[cfg(test)]
 mod tests {
-    use std::env;
-
     use super::*;
     use crate::color;
     use crate::tests::integration_test_utils;
 
+    // TODO: Test influence of BAT_THEME env var. E.g. see utils::process::tests::FakeParentArgs.
     #[test]
     fn test_syntax_theme_selection() {
         #[derive(PartialEq)]
@@ -134,52 +114,35 @@ mod tests {
         }
         for (
             syntax_theme,
-            bat_theme_env_var,
             mode, // (--light, --dark)
             expected_syntax_theme,
             expected_mode,
         ) in vec![
-            (None, "", None, DEFAULT_DARK_SYNTAX_THEME, Mode::Dark),
-            (Some("GitHub"), "", None, "GitHub", Mode::Light),
-            (Some("GitHub"), "1337", None, "GitHub", Mode::Light),
-            (None, "1337", None, "1337", Mode::Dark),
+            (None, None, DEFAULT_DARK_SYNTAX_THEME, Mode::Dark),
+            (Some("GitHub"), None, "GitHub", Mode::Light),
+            (Some("GitHub"), None, "GitHub", Mode::Light),
             (
                 None,
-                "<not set>",
-                None,
-                DEFAULT_DARK_SYNTAX_THEME,
-                Mode::Dark,
-            ),
-            (
-                None,
-                "",
                 Some(Mode::Light),
                 DEFAULT_LIGHT_SYNTAX_THEME,
                 Mode::Light,
             ),
             (
                 None,
-                "",
                 Some(Mode::Dark),
                 DEFAULT_DARK_SYNTAX_THEME,
                 Mode::Dark,
             ),
             (
                 None,
-                "<@@@@@>",
                 Some(Mode::Light),
                 DEFAULT_LIGHT_SYNTAX_THEME,
                 Mode::Light,
             ),
-            (None, "GitHub", Some(Mode::Light), "GitHub", Mode::Light),
-            (Some("none"), "", None, "none", Mode::Dark),
-            (Some("None"), "", Some(Mode::Light), "none", Mode::Light),
+            (None, Some(Mode::Light), "GitHub", Mode::Light),
+            (Some("none"), None, "none", Mode::Dark),
+            (Some("None"), Some(Mode::Light), "none", Mode::Light),
         ] {
-            if bat_theme_env_var == "<not set>" {
-                env::remove_var("BAT_THEME")
-            } else {
-                env::set_var("BAT_THEME", bat_theme_env_var);
-            }
             let mut args = vec![];
             if let Some(syntax_theme) = syntax_theme {
                 args.push("--syntax-theme");
