@@ -1,9 +1,11 @@
-use std::borrow::Cow;
 use std::collections::{HashMap, HashSet};
 use std::path::Path;
-use sysinfo::{Pid, Process, ProcessExt, SystemExt};
+use std::sync::{Arc, Condvar, Mutex, MutexGuard};
 
 use lazy_static::lazy_static;
+use sysinfo::{Pid, PidExt, Process, ProcessExt, ProcessRefreshKind, SystemExt};
+
+pub type DeltaPid = u32;
 
 #[derive(Clone, Debug, PartialEq)]
 pub enum CallingProcess {
@@ -13,9 +15,23 @@ pub enum CallingProcess {
     GitReflog(CommandLine),
     GitGrep(CommandLine),
     OtherGrep, // rg, grep, ag, ack, etc
+    None,      // no matching process could be found
+    Pending,   // calling process is currently being determined
 }
 // TODO: Git blame is currently handled differently
 
+impl CallingProcess {
+    pub fn paths_in_input_are_relative_to_cwd(&self) -> bool {
+        match self {
+            CallingProcess::GitDiff(cmd) if cmd.long_options.contains("--relative") => true,
+            CallingProcess::GitShow(cmd, _) if cmd.long_options.contains("--relative") => true,
+            CallingProcess::GitLog(cmd) if cmd.long_options.contains("--relative") => true,
+            CallingProcess::GitGrep(_) | CallingProcess::OtherGrep => true,
+            _ => false,
+        }
+    }
+}
+
 #[derive(Clone, Debug, PartialEq)]
 pub struct CommandLine {
     pub long_options: HashSet<String>,
@@ -23,23 +39,63 @@ pub struct CommandLine {
     last_arg: Option<String>,
 }
 
-pub fn calling_process() -> Option<Cow<'static, CallingProcess>> {
-    #[cfg(not(test))]
-    {
-        CACHED_CALLING_PROCESS.as_ref().map(Cow::Borrowed)
-    }
-    #[cfg(test)]
-    {
-        determine_calling_process().map(Cow::Owned)
-    }
-}
-
 lazy_static! {
-    static ref CACHED_CALLING_PROCESS: Option<CallingProcess> = determine_calling_process();
+    static ref CALLER: Arc<(Mutex<CallingProcess>, Condvar)> =
+        Arc::new((Mutex::new(CallingProcess::Pending), Condvar::new()));
 }
 
-fn determine_calling_process() -> Option<CallingProcess> {
+pub fn start_determining_calling_process_in_thread() {
+    // The handle is neither kept nor returned nor joined but dropped, so the main
+    // thread can exit early if it does not need to know its parent process.
+    std::thread::Builder::new()
+        .name("find_calling_process".into())
+        .spawn(move || {
+            let calling_process = determine_calling_process();
+
+            let (caller_mutex, determine_done) = &**CALLER;
+
+            let mut caller = caller_mutex.lock().unwrap();
+            *caller = calling_process;
+            determine_done.notify_all();
+        })
+        .unwrap();
+}
+
+#[cfg(not(test))]
+pub fn calling_process() -> MutexGuard<'static, CallingProcess> {
+    let (caller_mutex, determine_done) = &**CALLER;
+
+    determine_done
+        .wait_while(caller_mutex.lock().unwrap(), |caller| {
+            *caller == CallingProcess::Pending
+        })
+        .unwrap()
+}
+
+// The return value is duck-typed to work in place of a MutexGuard when testing.
+#[cfg(test)]
+pub fn calling_process() -> Box<CallingProcess> {
+    type _UnusedImport = MutexGuard<'static, i8>;
+
+    if crate::utils::process::tests::FakeParentArgs::are_set() {
+        // If the (thread-local) FakeParentArgs are set, then the following command returns
+        // these, so the cached global real ones can not be used.
+        Box::new(determine_calling_process())
+    } else {
+        let (caller_mutex, _) = &**CALLER;
+
+        let mut caller = caller_mutex.lock().unwrap();
+        if *caller == CallingProcess::Pending {
+            *caller = determine_calling_process();
+        }
+
+        Box::new(caller.clone())
+    }
+}
+
+fn determine_calling_process() -> CallingProcess {
     calling_process_cmdline(ProcInfo::new(), describe_calling_process)
+        .unwrap_or(CallingProcess::None)
 }
 
 // Return value of `extract_args(args: &[String]) -> ProcessArgs<T>` function which is
@@ -226,6 +282,13 @@ struct ProcInfo {
 }
 impl ProcInfo {
     fn new() -> Self {
+        // On Linux sysinfo optimizes for repeated process queries and keeps per-process
+        // /proc file descriptors open. This caching is not needed here, so
+        // set this to zero (this does nothing on other platforms).
+        // Also, there is currently a kernel bug which slows down syscalls when threads are
+        // involved (here: the ctrlc handler) and a lot of files are kept open.
+        sysinfo::set_open_files_limit(0);
+
         ProcInfo {
             info: sysinfo::System::new(),
         }
@@ -234,7 +297,8 @@ impl ProcInfo {
 
 trait ProcActions {
     fn cmd(&self) -> &[String];
-    fn parent(&self) -> Option<Pid>;
+    fn parent(&self) -> Option<DeltaPid>;
+    fn pid(&self) -> DeltaPid;
     fn start_time(&self) -> u64;
 }
 
@@ -245,8 +309,11 @@ where
     fn cmd(&self) -> &[String] {
         ProcessExt::cmd(self)
     }
-    fn parent(&self) -> Option<Pid> {
-        ProcessExt::parent(self)
+    fn parent(&self) -> Option<DeltaPid> {
+        ProcessExt::parent(self).map(|p| p.as_u32())
+    }
+    fn pid(&self) -> DeltaPid {
+        ProcessExt::pid(self).as_u32()
     }
     fn start_time(&self) -> u64 {
         ProcessExt::start_time(self)
@@ -256,35 +323,34 @@ where
 trait ProcessInterface {
     type Out: ProcActions;
 
-    fn my_pid(&self) -> Pid;
+    fn my_pid(&self) -> DeltaPid;
 
-    fn process(&self, pid: Pid) -> Option<&Self::Out>;
+    fn process(&self, pid: DeltaPid) -> Option<&Self::Out>;
     fn processes(&self) -> &HashMap<Pid, Self::Out>;
 
-    fn refresh_process(&mut self, pid: Pid) -> bool;
+    fn refresh_process(&mut self, pid: DeltaPid) -> bool;
     fn refresh_processes(&mut self);
 
-    fn parent_process(&mut self, pid: Pid) -> Option<&Self::Out> {
+    fn parent_process(&mut self, pid: DeltaPid) -> Option<&Self::Out> {
         self.refresh_process(pid).then(|| ())?;
         let parent_pid = self.process(pid)?.parent()?;
         self.refresh_process(parent_pid).then(|| ())?;
         self.process(parent_pid)
     }
-    fn naive_sibling_process(&mut self, pid: Pid) -> Option<&Self::Out> {
+    fn naive_sibling_process(&mut self, pid: DeltaPid) -> Option<&Self::Out> {
         let sibling_pid = pid - 1;
         self.refresh_process(sibling_pid).then(|| ())?;
         self.process(sibling_pid)
     }
-    #[cfg(test)]
-    fn find_sibling_process<F, T>(&mut self, pid: Pid, extract_args: F) -> Option<T>
+    fn find_sibling_in_refreshed_processes<F, T>(
+        &mut self,
+        pid: DeltaPid,
+        extract_args: &F,
+    ) -> Option<T>
     where
         F: Fn(&[String]) -> ProcessArgs<T>,
         Self: Sized,
     {
-        self.refresh_processes();
-
-        let this_start_time = self.process(pid)?.start_time();
-
         /*
 
         $ start_blame_of.sh src/main.rs | delta
@@ -303,7 +369,9 @@ trait ProcessInterface {
 
         */
 
-        let mut pid_distances = HashMap::<Pid, usize>::new();
+        let this_start_time = self.process(pid)?.start_time();
+
+        let mut pid_distances = HashMap::<DeltaPid, usize>::new();
         let mut collect_parent_pids = |pid, distance| {
             pid_distances.insert(pid, distance);
         };
@@ -330,9 +398,13 @@ trait ProcessInterface {
                             }
                         }
                     };
-                    iter_parents(self, pid, &mut sum_distance);
+                    iter_parents(self, pid.as_u32(), &mut sum_distance);
 
-                    Some((length_of_process_chain, args))
+                    if length_of_process_chain == usize::MAX {
+                        None
+                    } else {
+                        Some((length_of_process_chain, args))
+                    }
                 }
                 _ => None,
             })
@@ -346,20 +418,22 @@ trait ProcessInterface {
 impl ProcessInterface for ProcInfo {
     type Out = Process;
 
-    fn my_pid(&self) -> Pid {
-        std::process::id() as Pid
+    fn my_pid(&self) -> DeltaPid {
+        std::process::id()
     }
-    fn refresh_process(&mut self, pid: Pid) -> bool {
-        self.info.refresh_process(pid)
+    fn refresh_process(&mut self, pid: DeltaPid) -> bool {
+        self.info
+            .refresh_process_specifics(Pid::from_u32(pid), ProcessRefreshKind::new())
     }
-    fn process(&self, pid: Pid) -> Option<&Self::Out> {
-        self.info.process(pid)
+    fn process(&self, pid: DeltaPid) -> Option<&Self::Out> {
+        self.info.process(Pid::from_u32(pid))
     }
     fn processes(&self) -> &HashMap<Pid, Self::Out> {
         self.info.processes()
     }
     fn refresh_processes(&mut self) {
-        self.info.refresh_processes()
+        self.info
+            .refresh_processes_specifics(ProcessRefreshKind::new())
     }
 }
 
@@ -380,34 +454,47 @@ where
 
     let my_pid = info.my_pid();
 
-    // 1) Try the parent process. If delta is set as the pager in git, then git is the parent process.
-    let parent = info.parent_process(my_pid)?;
+    // 1) Try the parent process(es). If delta is set as the pager in git, then git is the parent process.
+    // If delta is started by a script check the parent's parent as well.
+    let mut current_pid = my_pid;
+    'parent_iter: for depth in [1, 2, 3] {
+        let parent = match info.parent_process(current_pid) {
+            None => {
+                break 'parent_iter;
+            }
+            Some(parent) => parent,
+        };
+        let parent_pid = parent.pid();
 
-    match extract_args(parent.cmd()) {
-        ProcessArgs::Args(result) => return Some(result),
-        ProcessArgs::ArgError => return None,
+        match extract_args(parent.cmd()) {
+            ProcessArgs::Args(result) => return Some(result),
+            ProcessArgs::ArgError => return None,
 
-        // 2) The parent process was something else, this can happen if git output is piped into delta, e.g.
-        // `git blame foo.txt | delta`. When the shell sets up the pipe it creates the two processes, the pids
-        // are usually consecutive, so check if the process with `my_pid - 1` matches.
-        ProcessArgs::OtherProcess => {
-            let sibling = info.naive_sibling_process(my_pid);
-            if let Some(proc) = sibling {
-                if let ProcessArgs::Args(result) = extract_args(proc.cmd()) {
-                    return Some(result);
+            // 2) The 1st parent process was something else, this can happen if git output is piped into delta, e.g.
+            // `git blame foo.txt | delta`. When the shell sets up the pipe it creates the two processes, the pids
+            // are usually consecutive, so naively check if the process with `my_pid - 1` matches.
+            ProcessArgs::OtherProcess if depth == 1 => {
+                let sibling = info.naive_sibling_process(current_pid);
+                if let Some(proc) = sibling {
+                    if let ProcessArgs::Args(result) = extract_args(proc.cmd()) {
+                        return Some(result);
+                    }
                 }
             }
-            // else try the fallback
+            // This check is not done for the parent's parent etc.
+            ProcessArgs::OtherProcess => {}
         }
+        current_pid = parent_pid;
     }
 
     /*
-    3) Neither parent nor direct sibling were a match.
+    3) Neither parent(s) nor the direct sibling were a match.
     The most likely case is that the input program of the pipe wrote all its data and exited before delta
     started, so no command line can be parsed. Same if the data was piped from an input file.
 
     There might also be intermediary scripts in between or piped input with a gap in pids or (rarely)
-    randomized pids, so check all processes for the closest match in the process tree.
+    randomized pids, so check processes for the closest match in the process tree.
+    The size of this process tree can be reduced by only refreshing selected processes.
 
     100 /usr/bin/some-terminal-emulator
     124  \_ -shell
@@ -426,28 +513,51 @@ where
     567  |       \_ less --RAW-CONTROL-CHARS --quit-if-one-screen
 
     */
-    #[cfg(test)]
-    {
-        info.find_sibling_process(my_pid, extract_args)
+
+    // Also `add` because `A_has_pid101 | delta_has_pid102`, but if A is a wrapper which then calls
+    // git (no `exec`), then the final pid of the git process might be 103 or greater.
+    let pid_range = my_pid.saturating_sub(10)..my_pid.saturating_add(10);
+    for p in pid_range {
+        // Processes which were not refreshed do not exist for sysinfo, so by selectively
+        // letting it know about processes the `find_sibling..` function will only
+        // consider these.
+        if info.process(p).is_none() {
+            info.refresh_process(p);
+        }
     }
-    #[cfg(not(test))]
-    {
-        None
+
+    match info.find_sibling_in_refreshed_processes(my_pid, &extract_args) {
+        None => {
+            #[cfg(not(target_os = "linux"))]
+            let full_scan = true;
+
+            // The full scan is expensive on Linux and rarely successful, so disable it by default.
+            #[cfg(target_os = "linux")]
+            let full_scan = std::env::var("DELTA_CALLING_PROCESS_QUERY_ALL")
+                .map_or(false, |v| !["0", "false", "no"].iter().any(|&n| n == v));
+
+            if full_scan {
+                info.refresh_processes();
+                info.find_sibling_in_refreshed_processes(my_pid, &extract_args)
+            } else {
+                None
+            }
+        }
+        some => some,
     }
 }
 
 // Walk up the process tree, calling `f` with the pid and the distance to `starting_pid`.
 // Prerequisite: `info.refresh_processes()` has been called.
-#[cfg(test)]
-fn iter_parents<P, F>(info: &P, starting_pid: Pid, f: F)
+fn iter_parents<P, F>(info: &P, starting_pid: DeltaPid, f: F)
 where
     P: ProcessInterface,
-    F: FnMut(Pid, usize),
+    F: FnMut(DeltaPid, usize),
 {
-    fn inner_iter_parents<P, F>(info: &P, pid: Pid, mut f: F, distance: usize)
+    fn inner_iter_parents<P, F>(info: &P, pid: DeltaPid, mut f: F, distance: usize)
     where
         P: ProcessInterface,
-        F: FnMut(Pid, usize),
+        F: FnMut(u32, usize),
     {
         // Probably bad input, not a tree:
         if distance > 2000 {
@@ -543,6 +653,9 @@ pub mod tests {
                 }
             })
         }
+        pub fn are_set() -> bool {
+            FAKE_ARGS.with(|a| *a.borrow() != TlsState::None)
+        }
         fn error(where_: &str) -> ! {
             panic!(
                 "test logic error (in {}): wrong FakeParentArgs scope?",
@@ -633,16 +746,26 @@ pub mod tests {
         assert_eq!(guess_git_blame_filename_extension(&args), Args("".into()));
     }
 
-    #[derive(Debug, Default)]
+    #[derive(Debug)]
     struct FakeProc {
         #[allow(dead_code)]
-        pid: Pid,
+        pid: DeltaPid,
         start_time: u64,
         cmd: Vec<String>,
-        ppid: Option<Pid>,
+        ppid: Option<DeltaPid>,
+    }
+    impl Default for FakeProc {
+        fn default() -> Self {
+            Self {
+                pid: 0,
+                start_time: 0,
+                cmd: Vec::new(),
+                ppid: None,
+            }
+        }
     }
     impl FakeProc {
-        fn new(pid: Pid, start_time: u64, cmd: Vec<String>, ppid: Option<Pid>) -> Self {
+        fn new(pid: DeltaPid, start_time: u64, cmd: Vec<String>, ppid: Option<DeltaPid>) -> Self {
             FakeProc {
                 pid,
                 start_time,
@@ -656,28 +779,42 @@ pub mod tests {
         fn cmd(&self) -> &[String] {
             &self.cmd
         }
-        fn parent(&self) -> Option<Pid> {
+        fn parent(&self) -> Option<DeltaPid> {
             self.ppid
         }
+        fn pid(&self) -> DeltaPid {
+            self.pid
+        }
         fn start_time(&self) -> u64 {
             self.start_time
         }
     }
 
-    #[derive(Debug, Default)]
+    #[derive(Debug)]
     struct MockProcInfo {
-        delta_pid: Pid,
+        delta_pid: DeltaPid,
         info: HashMap<Pid, FakeProc>,
     }
+    impl Default for MockProcInfo {
+        fn default() -> Self {
+            Self {
+                delta_pid: 0,
+                info: HashMap::new(),
+            }
+        }
+    }
     impl MockProcInfo {
-        fn with(processes: &[(Pid, u64, &str, Option<Pid>)]) -> Self {
+        fn with(processes: &[(DeltaPid, u64, &str, Option<DeltaPid>)]) -> Self {
             MockProcInfo {
                 delta_pid: processes.last().map(|p| p.0).unwrap_or(1),
                 info: processes
                     .iter()
                     .map(|(pid, start_time, cmd, ppid)| {
                         let cmd_vec = cmd.split(' ').map(str::to_owned).collect();
-                        (*pid, FakeProc::new(*pid, *start_time, cmd_vec, *ppid))
+                        (
+                            Pid::from_u32(*pid),
+                            FakeProc::new(*pid, *start_time, cmd_vec, *ppid),
+                        )
                     })
                     .collect(),
             }
@@ -687,17 +824,17 @@ pub mod tests {
     impl ProcessInterface for MockProcInfo {
         type Out = FakeProc;
 
-        fn my_pid(&self) -> Pid {
+        fn my_pid(&self) -> DeltaPid {
             self.delta_pid
         }
-        fn process(&self, pid: Pid) -> Option<&Self::Out> {
-            self.info.get(&pid)
+        fn process(&self, pid: DeltaPid) -> Option<&Self::Out> {
+            self.info.get(&Pid::from_u32(pid))
         }
         fn processes(&self) -> &HashMap<Pid, Self::Out> {
             &self.info
         }
         fn refresh_processes(&mut self) {}
-        fn refresh_process(&mut self, _pid: Pid) -> bool {
+        fn refresh_process(&mut self, _pid: DeltaPid) -> bool {
             true
         }
     }
@@ -806,6 +943,20 @@ pub mod tests {
         calling_process_cmdline(ProcInfo::new(), guess_git_blame_filename_extension);
     }
 
+    #[test]
+    fn test_process_blame_no_parent_found() {
+        let two_trees = MockProcInfo::with(&[
+            (2, 100, "-shell", None),
+            (3, 100, "git blame src/main.rs", Some(2)),
+            (4, 100, "call_delta.sh", None),
+            (5, 100, "delta", Some(4)),
+        ]);
+        assert_eq!(
+            calling_process_cmdline(two_trees, guess_git_blame_filename_extension),
+            None
+        );
+    }
+
     #[test]
     fn test_process_blame_info_with_parent() {
         let no_processes = MockProcInfo::with(&[]);
@@ -1028,7 +1179,7 @@ pub mod tests {
         info.refresh_processes();
         let mut ppid_distance = Vec::new();
 
-        iter_parents(&info, std::process::id() as Pid, |pid, distance| {
+        iter_parents(&info, std::process::id(), |pid, distance| {
             ppid_distance.push(pid as i32);
             ppid_distance.push(distance as i32)
         });
